#!/usr/bin/env python
# vim: et ts=4 sw=4


from django.template.loader import render_to_string
from .metatable import MetaTable
from .urls import extract, build


class Table(object):
    __metaclass__ = MetaTable

    def __init__(self, object_list=None, request=None, **kwargs):
        self._object_list = object_list
        self._request = request
        self._paginator = None

        if request is not None:
            prefix = kwargs.get('prefix', "")
            kwargs = dict(extract(request.GET, prefix), **kwargs)

        if len(kwargs):
            self._meta = self._meta.fork(
                **kwargs )

    def get_url(self, **kwargs):
        """
        Return an url, relative to the request associated with this
        table. Any keywords arguments provided added to the query
        string, replacing existing values.
        """

        return build(
            self._request.path,
            self._request.GET,
            self._meta.prefix,
            **kwargs )

    @property
    def object_list(self):
        """
        Return this table's object_list, transformed (sorted, reversed,
        filtered, etc) according to its meta options.
        """

        ol = self._object_list

        if self._meta.order_by and hasattr(ol, "order_by"):
            ol = ol.order_by(self._meta.order_by)

        return ol

    def as_html(self): # pragma: no cover
        """
        Return this table as HTML, ready to be inserted into a template.
        To customize the output, set the ``template`` meta option to a
        new template, and use the table_tags to build your own template.
        See README for examples.
        """

        return render_to_string(
            self._meta.template,
            { "table": self } )

    @property
    def paginator(self):
        if self._paginator is None:
            self._paginator = self._meta.paginator_class(
                self.object_list, self._meta.per_page )

        return self._paginator

    @property
    def columns(self):
        """Return the list of columns."""
        return self._meta.columns

    @property
    def rows(self):
        """Return the list of object on the active page."""

        return map(
            lambda o: self._meta.row_class(self, o),
            self.paginator.page(self._meta.page).object_list )

    def cell(self, column, row):
        return self._meta.cell_class(
            column, row )



#!/usr/bin/env python
# vim: et ts=4 sw=4


class Row(object):
    def __init__(self, table, obj):
        self.table = table
        self.obj = obj

    def __getattr__(self, name):
        if hasattr(self.obj, name):
            val = getattr(self.obj, name)

        elif hasattr(self.obj, "__getitem__") and (name in self.obj):
            val = self.obj[name]

        else:
            val = None

        return callable(val) and val() or val

    def __unicode__(self):
        return unicode(self.obj)

    def __iter__(self):
        for column in self.table._meta.columns:
            yield self.table.cell(column, self)

    def __len__(self):
        return len(self.table._meta.columns)



#!/usr/bin/env python
# vim: et ts=4 sw=4


import datetime
from django.template import defaultfilters


class Column(object):

    """
    This class represents a table column. It is responsible for holding
    metadata, and rending table cells. Like Django model/fields, columns
    are usually created within the table class which they are bound to.
    """

    creation_counter = 0

    def __init__(self, name=None, value=None, link=None, sortable=True):
        self._name = name
        self._value = value
        self._link = link
        self.sortable = sortable

        # like django fields, keep track of the order which columns are
        # created, so they can be sorted later. (unfortunately, python
        # attrs are passed to __new__ as an unsorted dict, so we must
        # keep track of this manually to avoid random column order.)
        self.creation_counter = Column.creation_counter
        Column.creation_counter += 1

        self.bound_to = None

    def __lt__(self, other):
        """Allow columns to be sorted by order of creation."""
        return self.creation_counter < other.creation_counter

    def __unicode__(self):
        return self.name

    def bind_to(self, table, name):
        """
        Bind this column to a table, and assign it a name. This method
        can only be called once per instance, because a Column cannot be
        bound to multiple tables. (The sort order would be ambiguous.)
        """

        if self.bound_to is not None:
            raise AttributeError(
                "Column is already bound to '%s' as '%s'" %\
                    self.bound_to)

        self.bound_to = (table, name)

    @property
    def is_bound(self):
        """Return true if this column is bound to a table."""
        return (self.bound_to is not None)

    @property
    def name(self):
        """Return the column name, whether explicit or implicit."""
        return self._name or self.bound_to[1]

    def value(self, cell):
        """
        Extract the value of ``cell``, ready to be rendered.

        If this Column was instantiated with a ``value`` attribute, it
        is called here to provide the value. (For example, to provide a
        calculated value.) Otherwise, ``cell.value`` is returned.
        """

        if self._value is not None:
            return self._value(cell)

        else:
            return cell.value

    def render(self, cell):
        """
        Render ``cell``, ready for display. The default behavior is to
        simply cast its value to unicode, but this may be overridden by
        child classes to do something more useful.
        """

        return unicode(self.value(cell))

    @property
    def has_link(self):
        """Return True if this column contains links."""
        return self._link is not None

    def link(self, cell):
        """
        Return the URL which ``cell`` should link to, or None if this
        column does not contain links.

        If this Column was instantiated with a ``link`` attribute, it is
        called here (with a single parameter of ``cell``), to provide
        the value. Otherwise, None is returned.
        """

        if self.has_link:
            return self._link(cell)

        return None


class DateColumn(Column):

    """
    This class provides a simple way to render a Date field, using the
    Django 'date' template filter. The ``format`` argument specifies the
    string in ``Django date format``_, **not** ``Python date format``_.
    If ``format`` is None the ``DATE_FORMAT`` setting is used.

    .. `Django date format``: http://docs.djangoproject.com/en/dev/ref/templates/builtins/#ttag-now
    .. `Python date format``: http://docs.python.org/library/datetime.html#strftime-strptime-behavior
    """

    def __init__(self, format=None, *args, **kwargs):
        super(DateColumn, self).__init__(*args, **kwargs)
        self._format = format

    def render(self, cell):
        return defaultfilters.date(
            self.value(cell),
            self._format)


class WrappedColumn(object):

    """
    This class wraps a Column instance, and binds it to a Table instance
    to provide useful properties to the template. This represents a per-
    render instance of the column, containing its volatile state, such
    as sort order. (More properties, such as visibility, filtering, and
    grouping may come later.)

    All of the attributes (and methods) of the wrapped Column can be
    accessed via this class, with help from some __getattr__ magic.
    """

    def __init__(self, table, column):
        self.table = table
        self.column = column

    @property
    def sort_url(self):
        """
        Return the URL to sort the linked table by this column. If the
        table is already sorted by this column, the order is reversed.

        Since there is no canonical URL for a table the current URL (via
        the HttpRequest linked to the Table instance) is reused, and any
        unrelated parameters will be included in the output.
        """

        prefix = (self.sort_direction == "asc") and "-" or ""
        return self.table.get_url(order_by=prefix + self.name)

    @property
    def is_sorted(self):
        return self.sort_direction is not None

    @property
    def sort_direction(self):
        """
        Return the direction in which the linked table is is sorted by
        this column ("asc" or "desc"), or None this column is unsorted.
        """

        if self.table._meta.order_by == self.name:
            return "asc"

        elif self.table._meta.order_by == ("-" + self.name):
            return "desc"

        else:
            return None

    def __unicode__(self):
        return unicode(self.column)

    def __getattr__(self, name):
        return getattr(self.column, name)



#!/usr/bin/env python
# vim: et ts=4 sw=4


from .options import TableOptions
from .column import Column


class MetaTable(type):
    def __new__(cls, name, bases, attrs):
        options_class = attrs.pop('options_class', TableOptions)
        column_class = attrs.pop('column_class', Column)

        # grab all of the instances of column_class from attrs, since
        # we're storing them separately (in _meta), we don't want them
        # in the actual table class.
        columns = dict([
            (attname, attrs.pop(attname))
            for attname, value in attrs.items()
            if isinstance(value, column_class)
        ])

        # create the class as usual (but sans attrs).
        attrs['_meta'] = options_class(attrs.pop('Meta', None))
        obj = super(MetaTable, cls).__new__(cls, name, bases, attrs)

        # bind each column, to make it aware of its own name. this
        # avoids having to include the name in the Column constructor.
        for name, column in columns.items():
            column.bind_to(obj, name)

        # store the columns (just the objects; they are already bound to
        # their names) in the order which they were created.
        obj._meta.columns = sorted(columns.values())

        return obj



#!/usr/bin/env python
# vim: et ts=4 sw=4


import copy


def extract(query_dict, prefix=""):
    """
    Extract the *order_by*, *per_page*, and *page* parameters from
    `query_dict` (a Django QueryDict), and return a dict suitable for
    instantiating a preconfigured Table object.
    """

    strs = ['order_by']
    ints = ['per_page', 'page']

    extracted = { }

    for key in (strs + ints):
        if (prefix + key) in query_dict:
            val = query_dict.get(prefix + key)

            extracted[key] = (val
                if not key in ints
                else int(val))

    return extracted


def build(path, query_dict, prefix="", **kwargs):
    """
    """

    query_dict = _copy(
        query_dict)

    for key in kwargs:
        query_dict[prefix + key] = kwargs[key]

    return "%s?%s" % (path, query_dict.urlencode())\
        if query_dict else path


def _copy(query_dict):
    """
    Return a mutable copy of `query_dict`. This is a workaround to
    Django bug #13572, which prevents QueryDict.copy from working.
    """

    memo = { }

    result = query_dict.__class__('',
        encoding=query_dict.encoding,
        mutable=True)

    memo[id(query_dict)] = result

    for key, value in dict.items(query_dict):
        dict.__setitem__(result,
            copy.deepcopy(key, memo),
            copy.deepcopy(value, memo))

    return result



#!/usr/bin/env python
# vim: et ts=4 sw=4


class Cell(object):

    """
    This class represents a single table cell. It does not hold a value,
    only references to the Column and Row which it intersects. The value
    is fetched from the Row, and rendered by the Column.
    """

    def __init__(self, column, row):
        self.column = column
        self.row = row

    def __unicode__(self):
        """Return the rendered (via self.column) value of this cell."""
        return unicode(self.column.render(self))

    @property
    def object(self):
        """
        Return the object (via self.row) which this cell contains. This
        property is included as a convenience, to make accessing the
        original object (which this Cell wraps) more natural.
        """

        return self.row.obj

    @property
    def value(self):
        """Return the value of this cell."""
        return getattr(self.row, self.column.name)

    @property
    def link(self):
        """Return the URL which this cell links to."""
        return self.column.link(self)



#!/usr/bin/env python
# vim: et ts=4 sw=4


from django import template
register = template.Library()

from ..column import WrappedColumn


@register.inclusion_tag("djtables/cols.html")
def table_cols(table):
    return {
        "columns": [
            WrappedColumn(table, column)
            for column in table.columns ] }


@register.inclusion_tag("djtables/head.html")
def table_head(table):
    return {
        "columns": [
            WrappedColumn(table, column)
            for column in table.columns ] }


@register.inclusion_tag("djtables/body.html")
def table_body(table):
    return {
        "rows": table.rows,
        "num_columns": len(table.columns) }


@register.inclusion_tag("djtables/foot.html")
def table_foot(table):
    return {
        "pages": [
            WrappedPage(table, number)
            for number in table.paginator.page_range ],
        "num_columns": len(table.columns) }


class WrappedPage(object):
    def __init__(self, table, number):
        self.table = table
        self.number = number

    @property
    def is_active(self):
        return self.table._meta.page == self.number

    def url(self):
        return self.table.get_url(page=self.number)



#!/usr/bin/env python
# vim: et ts=4 sw=4


from django.core.paginator import Paginator
from .row import Row
from .cell import Cell


class Options(object):
    _defaults = { }

    def __init__(self, options=None, **kwargs):

        # if an object was given, pluck our options from its attrs (like
        # django's confusingly-named Meta class). copy the attrs first,
        # to avoid trashing the object while we iterate it.
        options = (options is not None)\
            and options.__dict__.copy() or {}

        # keyword args are hard-coded options, so should override those
        # from the options object. (but they must still be valid.)
        options.update(kwargs)

        # store each option (except for _private and __magic__). setattr
        # will raise if ANY are invalid, like django. this seems kind of
        # heavy-handed (especially since i've wanted to add my own Meta
        # options to models before), but i'm following for consistency.
        for key in options.keys():
            if not key.startswith("_"):
                value = options.pop(key)
                setattr(self, key, value)

        # store any defaults which were not overridden.
        for key, value in self._defaults.items():
            if not hasattr(self, key):
                setattr(self, key, value)

    def __setattr__(self, name, value):
        if name in self._defaults:  object.__setattr__(self, name, value)
        else:  raise AttributeError("Invalid option: %s" % name)

    def fork(self, **kwargs):
        return self.__class__(self, **kwargs)


class TableOptions(Options):
    _defaults = {
        'paginator_class': Paginator,
        'row_class': Row,
        'cell_class': Cell,

        'prefix': "",
        'order_by': None,
        'per_page': 20,
        'page': 1,

        "template": "djtables/table.html",

        'columns': []
    }



#!/usr/bin/env python
# vim: et ts=4 sw=4


from setuptools import setup


setup(
    name="djtables",
    version="0.1.1",
    license="BSD",

    install_requires = [
        "django"
    ],

    package_dir={"": "lib"},
    packages=["djtables"],
    include_package_data=True,

    author="Adam Mckaig",
    author_email="adam.mckaig@gmail.com",

    description="Declarative HTML table builder for Django",
    url="http://github.com/adammck/djtables"
)
